

// start include version.h
#version 440 core
// end include version.h

layout (location = 0) in vec4  letterPos;
layout (location = 1) in float letterVal;
layout (location = 2) in vec4  letterXYZA;

out
// start include FONT_VS.h
FONT_VS
{
  vec2 fragmentPos;
  vec2 fletterPos;
  float fletterVal;
} font_vs;
// end include FONT_VS.h

uniform mat4 MVP;
uniform bool l3d = false;
uniform float length10;
uniform float scaleXYZ = 1.0f;
uniform float ratio = 1.0f;

// start include schmidt.h

uniform mat3 schmidt_rotd;
uniform mat3 schmidt_roti;
uniform float schmidt_opc2 = 1.25;
uniform float schmidt_omc2 = 0.75;
uniform bool schmidt_apply = false;

vec3 applySchmidt (vec3 pos)
{
  if (! schmidt_apply)
    return pos;

  vec3 XYZ = schmidt_roti * pos;
  XYZ = XYZ / length (XYZ);
  
  float lon = atan (XYZ.y, XYZ.x);
  float lat = asin (XYZ.z);
  
  float coordx = lon;
  float coordy = lat;

  float sincoordy = sin (coordy);

  float w = (schmidt_omc2 + sincoordy * schmidt_opc2) 
          / (schmidt_opc2 + sincoordy * schmidt_omc2);
  lat = asin (min (+1.0, max (-1.0, w)));
  lon = coordx;

  float coslat = cos (lat), sinlat = sin (lat);
  float coslon = cos (lon), sinlon = sin (lon);
  
  XYZ = vec3 (coslon * coslat, sinlon * coslat, sinlat);
  XYZ = schmidt_rotd * XYZ;
  XYZ = XYZ / length (XYZ);
  
  return XYZ;
}


// end include schmidt.h
// start include projection.h
// start include projection/kind.h
const int XYZ=0;
const int POLAR_NORTH=1;
const int POLAR_SOUTH=2;
const int MERCATOR=3;
const int LATLON=4;
// end include projection/kind.h

uniform int proj = 0;
uniform bool isflat = true;
uniform float lon0 = 180.0; // Longitude of right handside
const float pi = 3.1415926;

uniform bool apply_coordm = false;
uniform mat3 COORDM;

vec3 posFromLonLat (vec2 vertexLonLat)
{
  float lon = vertexLonLat.x, lat = vertexLonLat.y;
  if (abs (lat) > pi)
    return vec3 (0.0f, 0.0f, 0.0f);
  float coslon = cos (lon), sinlon = sin (lon);
  float coslat = cos (lat), sinlat = sin (lat);
  return vec3 (coslon * coslat, sinlon * coslat, sinlat);
}

vec3 compNormedPos (vec3 vertexPos)
{
  float x = vertexPos.x;
  float y = vertexPos.y;
  float z = vertexPos.z;
  float r = 1. / sqrt (x * x + y * y + z * z);
  return vec3 (x * r, y * r, z * r);
}

vec3 compProjedPos (vec3 vertexPos, vec3 normedPos)
{
  vec3 pos;
  normedPos = applySchmidt (normedPos);
  switch (proj)
    {
      case XYZ:
        if (isflat)
          pos = applySchmidt (normedPos);
        else
          pos = applySchmidt (vertexPos);
        break;
      case POLAR_NORTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (+p.z + 1.0),
                       p.y / (+p.z + 1.0));
        }
        break;
      case POLAR_SOUTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (-p.z + 1.0),
                       p.y / (-p.z + 1.0));
        }
        break;
      case MERCATOR:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = log (tan (pi / 4. + lat / 2.)) / pi;
          pos = vec3 (0., X, Y);
        }
        break;
      case LATLON:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = lat / pi;
          pos = vec3 (0., X, Y);
        }
        break;
    }

  return pos;
}

// end include projection.h
// start include scale.h
uniform float scale0 = 1.0f;

vec3 scalePosition (vec3 pos, vec3 normedPos, float scale0)
{
  normedPos = applySchmidt (normedPos);
  if (proj == XYZ)
    {
      pos.x = scale0 * normedPos.x;
      pos.y = scale0 * normedPos.y;
      pos.z = scale0 * normedPos.z;
    }
  else if (proj == POLAR_SOUTH)
    {
      pos.x = pos.x - (scale0 - 1.0f);
    }
  else
    {
      pos.x = pos.x + (scale0 - 1.0f);
    }

  return pos;
}

float scalingFactor (vec3 normedPos)
{
  if (proj == POLAR_SOUTH)
    {
      return 1.0 / (1.0 - normedPos.z);
    }
  else if (proj == POLAR_NORTH)
    {
      return 1.0 / (1.0 + normedPos.z);
    }
  else if (proj == MERCATOR)
    {
      return 1.0 / sqrt (1 - normedPos.z * normedPos.z);
    }
  return 1.0;
}

// end include scale.h

void main ()
{
  vec3 letterXYZ = letterXYZA.xyz;
  float A = letterXYZA.w;

  float xx = letterPos.x;
  float yy = letterPos.y;
  float dx = letterPos.z;
  float dy = letterPos.w;

  vec2 pos2 = vec2 (0.0f, 0.0f);

  if (gl_VertexID == 0)
    pos2 = vec2 (xx     , yy     );
  else if (gl_VertexID == 1)
    pos2 = vec2 (xx + dx, yy     );
  else if (gl_VertexID == 2)
    pos2 = vec2 (xx + dx, yy + dy);
  else if (gl_VertexID == 3)
    pos2 = vec2 (xx     , yy + dy);

  if (l3d) 
    {
      if (proj == XYZ)
        {
          float cosA = cos (A), sinA = sin (A);
          vec3 pos = scaleXYZ * letterXYZ;
          vec3 northPos  = vec3 (0., 0., 1.);
          vec3 vx = normalize (cross (northPos, pos));
          vec3 vy = normalize (cross (pos, vx));

          vec3 wx = + cosA * vx + sinA * vy;
          vec3 wy = - sinA * vx + cosA * vy;
         
          pos = pos + 50 * length10 * (pos2.x * wx + pos2.y * wy);
      
          gl_Position =  MVP * vec4 (pos, 1.);
        }
      else 
        {
          float scale0 = length (letterXYZ);
          vec3 vertexPos = letterXYZ;
          vec3 normedPos = compNormedPos (vertexPos);
          vec3 pos = compProjedPos (vertexPos, normedPos);
          pos = scalePosition (pos, normedPos, scale0);

          gl_Position =  MVP * vec4 (pos, 1.0f);

          if ((proj == POLAR_NORTH) || (proj == POLAR_SOUTH))
            {
              float p = proj == POLAR_NORTH ? +1.0f : -1.0f;
              vec3 posPole = scalePosition (vec3 (+0.0f, +0.0f, +0.0f), 
                                            vec3 (+0.0f, +0.0f,     p),
                                            scale0);
              vec4 pole_gl_Position = MVP * vec4 (posPole, 1.0f);
              A = A + pi / 2.0f + (p - 1.0f) * pi / 2.0f +
                  atan (gl_Position.y - pole_gl_Position.y, 
                        gl_Position.x - pole_gl_Position.x);
            }

          float cosA = cos (A), sinA = sin (A);

          gl_Position.x = gl_Position.x + 10. * (+ cosA * pos2.x - sinA * pos2.y) / ratio;
          gl_Position.y = gl_Position.y + 10. * (+ sinA * pos2.x + cosA * pos2.y);
          gl_Position.z = 0.0f;
        }
    }
  else 
    {
      float cosA = cos (A), sinA = sin (A);
      float Xr = letterXYZ.x, Yr = letterXYZ.y;
      vec2 rpos2 = vec2
      (
        Xr + (+ cosA * (pos2.x - Xr) - sinA * (pos2.y - Yr)),
        Yr + (+ sinA * (pos2.x - Xr) + cosA * (pos2.y - Yr)) 
      );
      gl_Position =  MVP * vec4 (0., rpos2.x, rpos2.y, 1.);
    }
  font_vs.fragmentPos = pos2;
  font_vs.fletterVal  = letterVal;
  font_vs.fletterPos  = vec2 (xx, yy);
}


