
// start include version.h
#version 440 core
// end include version.h

layout(location = 0) in vec3 vertexLonLatCst0;
layout(location = 1) in vec3 vertexLonLatCst1;
layout(location = 2) in vec3 vertexLonLatCst2;


out
// start include FRAME_VS.h
FRAME_VS
{
  float alpha;
  float islatcst;
  vec3 fragmentPos;
} frame_vs;
// end include FRAME_VS.h


uniform mat4 MVP;

// start include schmidt.h

uniform mat3 schmidt_rotd;
uniform mat3 schmidt_roti;
uniform float schmidt_opc2 = 1.25;
uniform float schmidt_omc2 = 0.75;
uniform bool schmidt_apply = false;

vec3 applySchmidt (vec3 pos)
{
  if (! schmidt_apply)
    return pos;

  vec3 XYZ = schmidt_roti * pos;
  XYZ = XYZ / length (XYZ);
  
  float lon = atan (XYZ.y, XYZ.x);
  float lat = asin (XYZ.z);
  
  float coordx = lon;
  float coordy = lat;

  float sincoordy = sin (coordy);

  float w = (schmidt_omc2 + sincoordy * schmidt_opc2) 
          / (schmidt_opc2 + sincoordy * schmidt_omc2);
  lat = asin (min (+1.0, max (-1.0, w)));
  lon = coordx;

  float coslat = cos (lat), sinlat = sin (lat);
  float coslon = cos (lon), sinlon = sin (lon);
  
  XYZ = vec3 (coslon * coslat, sinlon * coslat, sinlat);
  XYZ = schmidt_rotd * XYZ;
  XYZ = XYZ / length (XYZ);
  
  return XYZ;
}


// end include schmidt.h
// start include projection.h
// start include projection/kind.h
const int XYZ=0;
const int POLAR_NORTH=1;
const int POLAR_SOUTH=2;
const int MERCATOR=3;
const int LATLON=4;
// end include projection/kind.h

uniform int proj = 0;
uniform bool isflat = true;
uniform float lon0 = 180.0; // Longitude of right handside
const float pi = 3.1415926;

uniform bool apply_coordm = false;
uniform mat3 COORDM;

vec3 posFromLonLat (vec2 vertexLonLat)
{
  float lon = vertexLonLat.x, lat = vertexLonLat.y;
  if (abs (lat) > pi)
    return vec3 (0.0f, 0.0f, 0.0f);
  float coslon = cos (lon), sinlon = sin (lon);
  float coslat = cos (lat), sinlat = sin (lat);
  return vec3 (coslon * coslat, sinlon * coslat, sinlat);
}

vec3 compNormedPos (vec3 vertexPos)
{
  float x = vertexPos.x;
  float y = vertexPos.y;
  float z = vertexPos.z;
  float r = 1. / sqrt (x * x + y * y + z * z);
  return vec3 (x * r, y * r, z * r);
}

vec3 compProjedPos (vec3 vertexPos, vec3 normedPos)
{
  vec3 pos;
  normedPos = applySchmidt (normedPos);
  switch (proj)
    {
      case XYZ:
        if (isflat)
          pos = applySchmidt (normedPos);
        else
          pos = applySchmidt (vertexPos);
        break;
      case POLAR_NORTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (+p.z + 1.0),
                       p.y / (+p.z + 1.0));
        }
        break;
      case POLAR_SOUTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (-p.z + 1.0),
                       p.y / (-p.z + 1.0));
        }
        break;
      case MERCATOR:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = log (tan (pi / 4. + lat / 2.)) / pi;
          pos = vec3 (0., X, Y);
        }
        break;
      case LATLON:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = lat / pi;
          pos = vec3 (0., X, Y);
        }
        break;
    }

  return pos;
}

// end include projection.h
// start include scale.h
uniform float scale0 = 1.0f;

vec3 scalePosition (vec3 pos, vec3 normedPos, float scale0)
{
  normedPos = applySchmidt (normedPos);
  if (proj == XYZ)
    {
      pos.x = scale0 * normedPos.x;
      pos.y = scale0 * normedPos.y;
      pos.z = scale0 * normedPos.z;
    }
  else if (proj == POLAR_SOUTH)
    {
      pos.x = pos.x - (scale0 - 1.0f);
    }
  else
    {
      pos.x = pos.x + (scale0 - 1.0f);
    }

  return pos;
}

float scalingFactor (vec3 normedPos)
{
  if (proj == POLAR_SOUTH)
    {
      return 1.0 / (1.0 - normedPos.z);
    }
  else if (proj == POLAR_NORTH)
    {
      return 1.0 / (1.0 + normedPos.z);
    }
  else if (proj == MERCATOR)
    {
      return 1.0 / sqrt (1 - normedPos.z * normedPos.z);
    }
  return 1.0;
}

// end include scale.h

uniform bool do_alpha = false;
uniform float posmax = 0.97;
uniform float width = 0.005;

void main ()
{
  vec3 vertexPos0 = posFromLonLat (vertexLonLatCst0.xy);
  vec3 vertexPos1 = posFromLonLat (vertexLonLatCst1.xy);
  vec3 vertexPos2 = posFromLonLat (vertexLonLatCst2.xy);

  frame_vs.islatcst = vertexLonLatCst0.z;

  vec3 vertexPos;

  vec3 t0 = normalize (vertexPos1 - vertexPos0);
  vec3 t1 = normalize (vertexPos2 - vertexPos1);

  if ((gl_VertexID == 0) || (gl_VertexID == 2))
    {
      vertexPos = vertexPos0;
    }
  else if ((gl_VertexID == 1) || (gl_VertexID == 3) || (gl_VertexID == 5) || (gl_VertexID == 4))
    {
      vertexPos = vertexPos1;  
    }

  frame_vs.fragmentPos = vertexPos;

  vec3 vertexPosA = vertexPos;
  vec3 vertexPosB = vertexPos;

  vec3 p = normalize (vertexPosB);
  vec3 n0 = cross (t0, p);
  vec3 n1 = cross (t1, p);

  float c = width / scalingFactor (p);

  if ((gl_VertexID >= 4) && (dot (cross (n0, n1), vertexPosB) < 0.))
    c = 0.0;

  if (gl_VertexID == 2)
    vertexPosB = vertexPosB + c * n0;
  if (gl_VertexID == 3)
    vertexPosB = vertexPosB + c * n0;
  if (gl_VertexID == 4)
    vertexPosB = vertexPosB + c * normalize (n0 + n1);
  if (gl_VertexID == 5)
    vertexPosB = vertexPosB + c * n1;

  vec3 normedPosA = compNormedPos (vertexPosA);
  vec3 normedPosB = compNormedPos (vertexPosB);
  vec3 posA = compProjedPos (vertexPosA, normedPosA);
  vec3 posB = compProjedPos (vertexPosB, normedPosB);

  frame_vs.alpha = min (min (1.0f, length (vertexPos0)), min (1.0f, length (vertexPos1)));

  if (proj == XYZ)
    {
      posB = scalePosition (posB, normedPosB, scale0);
    }
  else 
    {
      if ((gl_VertexID != 0) && (gl_VertexID != 1))
        posB = posA + width * 0.3 * normalize (posB - posA);

      if ((proj == LATLON) || (proj == MERCATOR))
      if ((posB.y < -posmax) || (+posmax < posB.y))
        {
          posB.x = -0.1;
          if (do_alpha)
            frame_vs.alpha = 0.0;
	}
      if (proj == LATLON)
      if ((posB.z > +0.49) || (posB.z < -0.49))
        frame_vs.alpha = 0.0;

      if (proj == POLAR_SOUTH)
        posB.x = posB.x - 0.005;
      else
        posB.x = posB.x + 0.005;
    }

  gl_Position =  MVP * vec4 (posB, 1);

}
