
// start include version.h
#version 440 core
// end include version.h

layout (location = 0) in vec3 aPos;

uniform mat4 MVP;

// start include schmidt.h

uniform mat3 schmidt_rotd;
uniform mat3 schmidt_roti;
uniform float schmidt_opc2 = 1.25;
uniform float schmidt_omc2 = 0.75;
uniform bool schmidt_apply = false;

vec3 applySchmidt (vec3 pos)
{
  if (! schmidt_apply)
    return pos;

  vec3 XYZ = schmidt_roti * pos;
  XYZ = XYZ / length (XYZ);
  
  float lon = atan (XYZ.y, XYZ.x);
  float lat = asin (XYZ.z);
  
  float coordx = lon;
  float coordy = lat;

  float sincoordy = sin (coordy);

  float w = (schmidt_omc2 + sincoordy * schmidt_opc2) 
          / (schmidt_opc2 + sincoordy * schmidt_omc2);
  lat = asin (min (+1.0, max (-1.0, w)));
  lon = coordx;

  float coslat = cos (lat), sinlat = sin (lat);
  float coslon = cos (lon), sinlon = sin (lon);
  
  XYZ = vec3 (coslon * coslat, sinlon * coslat, sinlat);
  XYZ = schmidt_rotd * XYZ;
  XYZ = XYZ / length (XYZ);
  
  return XYZ;
}


// end include schmidt.h
// start include projection.h
// start include projection/kind.h
const int XYZ=0;
const int POLAR_NORTH=1;
const int POLAR_SOUTH=2;
const int MERCATOR=3;
const int LATLON=4;
// end include projection/kind.h

uniform int proj = 0;
uniform bool isflat = true;
uniform float lon0 = 180.0; // Longitude of right handside
const float pi = 3.1415926;

uniform bool apply_coordm = false;
uniform mat3 COORDM;

vec3 posFromLonLat (vec2 vertexLonLat)
{
  float lon = vertexLonLat.x, lat = vertexLonLat.y;
  if (abs (lat) > pi)
    return vec3 (0.0f, 0.0f, 0.0f);
  float coslon = cos (lon), sinlon = sin (lon);
  float coslat = cos (lat), sinlat = sin (lat);
  return vec3 (coslon * coslat, sinlon * coslat, sinlat);
}

vec3 compNormedPos (vec3 vertexPos)
{
  float x = vertexPos.x;
  float y = vertexPos.y;
  float z = vertexPos.z;
  float r = 1. / sqrt (x * x + y * y + z * z);
  return vec3 (x * r, y * r, z * r);
}

vec3 compProjedPos (vec3 vertexPos, vec3 normedPos)
{
  vec3 pos;
  normedPos = applySchmidt (normedPos);
  switch (proj)
    {
      case XYZ:
        if (isflat)
          pos = applySchmidt (normedPos);
        else
          pos = applySchmidt (vertexPos);
        break;
      case POLAR_NORTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (+p.z + 1.0),
                       p.y / (+p.z + 1.0));
        }
        break;
      case POLAR_SOUTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (-p.z + 1.0),
                       p.y / (-p.z + 1.0));
        }
        break;
      case MERCATOR:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = log (tan (pi / 4. + lat / 2.)) / pi;
          pos = vec3 (0., X, Y);
        }
        break;
      case LATLON:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = lat / pi;
          pos = vec3 (0., X, Y);
        }
        break;
    }

  return pos;
}

// end include projection.h
// start include scale.h
uniform float scale0 = 1.0f;

vec3 scalePosition (vec3 pos, vec3 normedPos, float scale0)
{
  normedPos = applySchmidt (normedPos);
  if (proj == XYZ)
    {
      pos.x = scale0 * normedPos.x;
      pos.y = scale0 * normedPos.y;
      pos.z = scale0 * normedPos.z;
    }
  else if (proj == POLAR_SOUTH)
    {
      pos.x = pos.x - (scale0 - 1.0f);
    }
  else
    {
      pos.x = pos.x + (scale0 - 1.0f);
    }

  return pos;
}

float scalingFactor (vec3 normedPos)
{
  if (proj == POLAR_SOUTH)
    {
      return 1.0 / (1.0 - normedPos.z);
    }
  else if (proj == POLAR_NORTH)
    {
      return 1.0 / (1.0 + normedPos.z);
    }
  else if (proj == MERCATOR)
    {
      return 1.0 / sqrt (1 - normedPos.z * normedPos.z);
    }
  return 1.0;
}

// end include scale.h

uniform float length10 = 0.01;
uniform float valmin;
uniform float valmax;
uniform float pointSiz;
uniform bool lpointSiz;
uniform bool lpointZoo = false;
uniform float ratio = 1.0f;

out float pointVal;
out vec3 centerVec;
flat out float pointRad;

void main()
{

  pointVal = aPos.z;

  vec2 pos2 = vec2 (0.0f, 0.0f);

  if (gl_VertexID == 0)
    pos2 = vec2 (-1.0f, -1.0f);
  else if (gl_VertexID == 1)
    pos2 = vec2 (+1.0f, -1.0f);
  else if (gl_VertexID == 2)
    pos2 = vec2 (+1.0f, +1.0f);
  else if (gl_VertexID == 3)
    pos2 = vec2 (-1.0f, +1.0f);
  
  float lon = aPos.x;
  float lat = aPos.y;
  float siz = 0.5 * pointSiz;
  if (lpointSiz)
    siz = siz * pointVal / valmax;
  
  vec3 pos = posFromLonLat (vec2 (lon, lat));

  if (proj == XYZ)
    {
      pos = applySchmidt (pos);
      pos = pos * scale0;

      vec3 northPos  = vec3 (0., 0., 1.);
      vec3 vx = normalize (cross (northPos, pos));
      vec3 vy = normalize (cross (pos, vx));

      if (lpointZoo)
        pointRad = siz * 0.02;
      else
        pointRad = siz * length10;

      centerVec = pointRad * (pos2.x * vx + pos2.y * vy); 
      pos = pos + centerVec;

      gl_Position = MVP * vec4 (pos, 1.);

    }
  else
    {
      float scale0 = length (pos);
      vec3 vertexPos = pos;
      vec3 normedPos = compNormedPos (vertexPos);
      vec3 pos = compProjedPos (vertexPos, normedPos);
      pos = scalePosition (pos, normedPos, scale0);

      gl_Position =  MVP * vec4 (pos, 1.);

      pointRad = siz * 0.1;
      centerVec.xy = pointRad * pos2;
      centerVec.z = 0.0f;

      gl_Position.x = gl_Position.x + centerVec.x / ratio;
      gl_Position.y = gl_Position.y + centerVec.y;
      gl_Position.z = 0.0f;


    }

}
