
#version 440 core

layout(location = 0) in vec2  vertexLonLat;
layout(location = 1) in float vertexVal;
layout(location = 2) in float vertexHeight;
layout(location = 3) in vec3  vertexMPIView;


out 
// start include SCALAR_VS.h
SCALAR_VS
{
  float fragmentVal;
  float fragmentMPI;
  vec3 fragmentPos;
  float missingFlag;
  flat float fragmentValFlat;
  flat float fragmentMPIFlat;
} scalar_vs;
// end include SCALAR_VS.h

uniform mat4 MVP;
uniform float height_scale = 0.05;
uniform float mpiview_scale = 0.0f;

// start include schmidt.h

uniform mat3 schmidt_rotd;
uniform mat3 schmidt_roti;
uniform float schmidt_opc2 = 1.25;
uniform float schmidt_omc2 = 0.75;
uniform bool schmidt_apply = false;

vec3 applySchmidt (vec3 pos)
{
  if (! schmidt_apply)
    return pos;

  vec3 XYZ = schmidt_roti * pos;
  XYZ = XYZ / length (XYZ);
  
  float lon = atan (XYZ.y, XYZ.x);
  float lat = asin (XYZ.z);
  
  float coordx = lon;
  float coordy = lat;

  float sincoordy = sin (coordy);

  float w = (schmidt_omc2 + sincoordy * schmidt_opc2) 
          / (schmidt_opc2 + sincoordy * schmidt_omc2);
  lat = asin (min (+1.0, max (-1.0, w)));
  lon = coordx;

  float coslat = cos (lat), sinlat = sin (lat);
  float coslon = cos (lon), sinlon = sin (lon);
  
  XYZ = vec3 (coslon * coslat, sinlon * coslat, sinlat);
  XYZ = schmidt_rotd * XYZ;
  XYZ = XYZ / length (XYZ);
  
  return XYZ;
}


// end include schmidt.h
// start include projection.h
// start include projection/kind.h
const int XYZ=0;
const int POLAR_NORTH=1;
const int POLAR_SOUTH=2;
const int MERCATOR=3;
const int LATLON=4;
// end include projection/kind.h

uniform int proj = 0;
uniform bool isflat = true;
uniform float lon0 = 180.0; // Longitude of right handside
const float pi = 3.1415926;

uniform bool apply_coordm = false;
uniform mat3 COORDM;

vec3 posFromLonLat (vec2 vertexLonLat)
{
  float lon = vertexLonLat.x, lat = vertexLonLat.y;
  if (abs (lat) > pi)
    return vec3 (0.0f, 0.0f, 0.0f);
  float coslon = cos (lon), sinlon = sin (lon);
  float coslat = cos (lat), sinlat = sin (lat);
  return vec3 (coslon * coslat, sinlon * coslat, sinlat);
}

vec3 compNormedPos (vec3 vertexPos)
{
  float x = vertexPos.x;
  float y = vertexPos.y;
  float z = vertexPos.z;
  float r = 1. / sqrt (x * x + y * y + z * z);
  return vec3 (x * r, y * r, z * r);
}

vec3 compProjedPos (vec3 vertexPos, vec3 normedPos)
{
  vec3 pos;
  normedPos = applySchmidt (normedPos);
  switch (proj)
    {
      case XYZ:
        if (isflat)
          pos = applySchmidt (normedPos);
        else
          pos = applySchmidt (vertexPos);
        break;
      case POLAR_NORTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (+p.z + 1.0),
                       p.y / (+p.z + 1.0));
        }
        break;
      case POLAR_SOUTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (-p.z + 1.0),
                       p.y / (-p.z + 1.0));
        }
        break;
      case MERCATOR:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = log (tan (pi / 4. + lat / 2.)) / pi;
          pos = vec3 (0., X, Y);
        }
        break;
      case LATLON:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = lat / pi;
          pos = vec3 (0., X, Y);
        }
        break;
    }

  return pos;
}

// end include projection.h
// start include scale.h
uniform float scale0 = 1.0f;

vec3 scalePosition (vec3 pos, vec3 normedPos, float scale0)
{
  normedPos = applySchmidt (normedPos);
  if (proj == XYZ)
    {
      pos.x = scale0 * normedPos.x;
      pos.y = scale0 * normedPos.y;
      pos.z = scale0 * normedPos.z;
    }
  else if (proj == POLAR_SOUTH)
    {
      pos.x = pos.x - (scale0 - 1.0f);
    }
  else
    {
      pos.x = pos.x + (scale0 - 1.0f);
    }

  return pos;
}

float scalingFactor (vec3 normedPos)
{
  if (proj == POLAR_SOUTH)
    {
      return 1.0 / (1.0 - normedPos.z);
    }
  else if (proj == POLAR_NORTH)
    {
      return 1.0 / (1.0 + normedPos.z);
    }
  else if (proj == MERCATOR)
    {
      return 1.0 / sqrt (1 - normedPos.z * normedPos.z);
    }
  return 1.0;
}

// end include scale.h

const float rad2deg = 180.0 / pi;
const float deg2rad = pi / 180.0;

// start include geometry.h

// start include geometry/buffer_index.h

const int geometry_gaussian_jlat_idx = 1;
const int geometry_gaussian_jglo_idx = 2;
const int geometry_gaussian_glat_idx = 3;


// end include geometry/buffer_index.h
// start include geometry/types.h
const int geometry_none     = 0;
const int geometry_gaussian = 1;
const int geometry_lambert  = 2;
const int geometry_latlon   = 3;

const int dumm_type = geometry_none + geometry_gaussian + geometry_lambert + geometry_latlon;
// end include geometry/types.h
// start include geometry/gaussian.h

// start include geometry/gaussian/kinds.h

const int geometry_gaussian_kind_unk = 0;
const int geometry_gaussian_kind_lin = 1;
const int geometry_gaussian_kind_oct = 2;

const int dumm_kind = geometry_gaussian_kind_unk + geometry_gaussian_kind_lin + geometry_gaussian_kind_oct;

// end include geometry/gaussian/kinds.h

layout (std430, binding=geometry_gaussian_jlat_idx) buffer geometry_gaussian1
{
  int geometry_gaussian_jlat[];
};

layout (std430, binding=geometry_gaussian_jglo_idx) buffer geometry_gaussian2
{
  int geometry_gaussian_jglooff[];
};

layout (std430, binding=geometry_gaussian_glat_idx) buffer geometry_gaussian3
{
  float geometry_gaussian_latgauss[];
};

uniform int   geometry_gaussian_Nj;
uniform mat3  geometry_gaussian_rot;
uniform bool  geometry_gaussian_rotated;
uniform float geometry_gaussian_omc2;
uniform float geometry_gaussian_opc2;
uniform float geometry_gaussian_latfit_coeff[10];
uniform int   geometry_gaussian_latfit_degre;
uniform int   geometry_gaussian_numberOfPoints;
uniform bool  geometry_gaussian_fitlat;
uniform int   geometry_gaussian_kind;

void geometry_gaussian_miss ()
{
}

// start include geometry/gaussian/guess_jlat.h
float geometry_gaussian_guess_lat_lin (int jglo)
{
  return asin (1.0 - 2.0 * float (jglo) / float (geometry_gaussian_numberOfPoints));
}

float geometry_gaussian_guess_lat_oct (int jglo)
{
  return 0.5 * pi * (1.0 - sqrt (2.0 * float (jglo) / float (geometry_gaussian_numberOfPoints)));
}

float geometry_gaussian_guess_lat (int jglo)
{
  float lat = 0.0;
  switch (geometry_gaussian_kind)
    {
      case geometry_gaussian_kind_lin:
        lat = geometry_gaussian_guess_lat_lin (jglo);
        break;
      case geometry_gaussian_kind_oct:
        lat = geometry_gaussian_guess_lat_oct (jglo);
        break;
    }
  return lat;
}


float geometry_gaussian_latfit_eval (int jlat)
{
  float x = float (jlat);
  if (geometry_gaussian_latfit_degre == 4)
    return geometry_gaussian_latfit_coeff[0] + x *
	  (geometry_gaussian_latfit_coeff[1] + x *
	  (geometry_gaussian_latfit_coeff[2] + x *
	  (geometry_gaussian_latfit_coeff[3] + x *
	  (geometry_gaussian_latfit_coeff[4]))));
  else if (geometry_gaussian_latfit_degre == 3)
    return geometry_gaussian_latfit_coeff[0] + x *
	  (geometry_gaussian_latfit_coeff[1] + x *
	  (geometry_gaussian_latfit_coeff[2] + x *
	  (geometry_gaussian_latfit_coeff[3])));
  float y = 0;
  for (int i = geometry_gaussian_latfit_degre; i >= 0; i--)
     y = y * x + geometry_gaussian_latfit_coeff[i];
  return y;
}

int geometry_gaussian_guess_jlat0 (int jglo)
{
  float lat = geometry_gaussian_guess_lat (jglo);
  return int ((0.5 - lat / pi) * (geometry_gaussian_Nj + 1) - 1);
}

int geometry_gaussian_guess_jlat (int jglo)
{
  bool south = jglo > geometry_gaussian_numberOfPoints / 2;
  int iglo = south ? geometry_gaussian_numberOfPoints - 1 - jglo : jglo;

  int ilat0 = geometry_gaussian_guess_jlat0 (iglo);

  // Correction with polynomial
  int ilat1 = int (geometry_gaussian_latfit_eval (ilat0));
  int jlat = ilat0 + ilat1;

  jlat = max (min (jlat, geometry_gaussian_Nj), 0);

  while (iglo < geometry_gaussian_jglooff[jlat])
    {
      jlat = jlat - 1;
      geometry_gaussian_miss ();
    }

  while (iglo >= geometry_gaussian_jglooff[jlat+1])
    {
      jlat = jlat + 1;
      geometry_gaussian_miss ();
    }

  jlat = south ? geometry_gaussian_Nj - 1 - jlat : jlat;

  return jlat;
}


// end include geometry/gaussian/guess_jlat.h

vec2 getGaussianVertexLonLat (int jglo) 
{
  const float twopi = 2.0f * pi;
  int jlat;

  if (geometry_gaussian_fitlat)
    {
      jlat = geometry_gaussian_guess_jlat (jglo);
    }
  else
    {
      jlat = geometry_gaussian_jlat[jglo];
    }

  int jlon = jglo - geometry_gaussian_jglooff[jlat];

  float coordy = geometry_gaussian_latgauss[jlat];
  int pl = geometry_gaussian_jglooff[jlat+1] - geometry_gaussian_jglooff[jlat];
  float coordx = (twopi * float (jlon)) / float (pl);

  float lon, lat;

  if (! geometry_gaussian_rotated)
    {
      lon = coordx;
      lat = coordy;
    }
  else
    {
      float sincoordy = sin (coordy);
      lat = asin ((geometry_gaussian_omc2 + sincoordy * geometry_gaussian_opc2) 
                / (geometry_gaussian_opc2 + sincoordy * geometry_gaussian_omc2));
      lon = coordx;

      float coslat = cos (lat), sinlat = sin (lat);
      float coslon = cos (lon), sinlon = sin (lon);
  
      float X = coslon * coslat;
      float Y = sinlon * coslat;
      float Z =          sinlat;
  
      vec3 XYZ = vec3 (X, Y, Z);
      XYZ = geometry_gaussian_rot * XYZ;
  
      lon = atan (XYZ.y, XYZ.x);
      lat = asin (XYZ.z);
    }


  return vec2 (lon, lat);
}

// end include geometry/gaussian.h
// start include geometry/lambert.h


float sign2 (float a, float b)
{
  return b > 0 ? +abs (a) : -abs (a);
}

struct latlon_t
{
  float lon, lat;
};

struct latlon_j_t
{
  latlon_t V, A, B;
};

struct xy_t
{
  float x, y;
};

struct xy_j_t
{
  xy_t V, A, B;
};

struct rtheta_t
{
  float r, theta;
};

struct rtheta_j_t
{
  rtheta_t V, A, B;
};

float dist_2ref (latlon_t pt_coord, latlon_t ref_coord)
{
  float z = pt_coord.lon - ref_coord.lon;
  z = z - sign2 (pi, z) * (1.0 + sign2 (1.0, abs (z) - pi));
  return -z * sign2 (1.0, z - pi);
}
  
struct proj_t
{
  latlon_t ref_pt;
  float pole;
  float r_equateur;
  float kl;
};

uniform float geometry_lambert_p_pj_ref_pt_lon;
uniform float geometry_lambert_p_pj_ref_pt_lat;
uniform float geometry_lambert_p_pj_pole;
uniform float geometry_lambert_p_pj_r_equateur;
uniform float geometry_lambert_p_pj_kl;
uniform int   geometry_lambert_Nx;
uniform int   geometry_lambert_Ny;
uniform int   geometry_lambert_Nux;
uniform int   geometry_lambert_Nuy;
uniform float geometry_lambert_DxInMetres;
uniform float geometry_lambert_DyInMetres;
uniform float geometry_lambert_center_xy_x;
uniform float geometry_lambert_center_xy_y;

proj_t get_proj_t ()
{
  proj_t proj = proj_t (latlon_t (geometry_lambert_p_pj_ref_pt_lon,
                                  geometry_lambert_p_pj_ref_pt_lat),
                        geometry_lambert_p_pj_pole,
                        geometry_lambert_p_pj_r_equateur,
                        geometry_lambert_p_pj_kl);
  return proj;
}

rtheta_t xy_to_rtheta (proj_t proj, xy_t pt_xy)
{
  float r = sqrt (pt_xy.x * pt_xy.x + pt_xy.y * pt_xy.y);
  float tatng;
  float theta;

  if (pt_xy.y == 0.0) 
    tatng = (pt_xy.x == 0.0) ? pi : sign2 (pi / 2.0, -proj.pole * pt_xy.x);
  else
    tatng = atan (-proj.pole * (pt_xy.x / pt_xy.y));
  theta = pi * sign2 (1.0, pt_xy.x) * (sign2 (0.5, proj.pole * pt_xy.y) + 0.5) + tatng;

  return rtheta_t (r, theta);
}

latlon_t rtheta_to_latlon (proj_t proj, rtheta_t pt_rtheta) 
{
  return latlon_t (proj.ref_pt.lon + pt_rtheta.theta / proj.kl,
                   proj.pole * ((pi / 2.0) - 2.0 * atan (pow (pt_rtheta.r / proj.r_equateur, 1.0 / proj.kl))));
}

latlon_t xy_to_latlon (proj_t proj, xy_t pt_xy) 
{
  return rtheta_to_latlon (proj, xy_to_rtheta (proj, pt_xy));
}

vec2 getLambertVertexLonLat (int jglo)
{
  int i = jglo % geometry_lambert_Nx;
  int j = jglo / geometry_lambert_Nx;

  proj_t proj = get_proj_t ();

  xy_t pt_xy = xy_t ((i - geometry_lambert_Nux / 2) * geometry_lambert_DxInMetres, 
                     (j - geometry_lambert_Nuy / 2) * geometry_lambert_DyInMetres);
  
  pt_xy.x = pt_xy.x + geometry_lambert_center_xy_x;
  pt_xy.y = pt_xy.y + geometry_lambert_center_xy_y;

  latlon_t latlon = xy_to_latlon (proj, pt_xy);

  return vec2 (latlon.lon, latlon.lat);
}  



// end include geometry/lambert.h
// start include geometry/latlon.h
uniform float geometry_latlon_lat0;
uniform float geometry_latlon_lon0;
uniform float geometry_latlon_dlat;
uniform float geometry_latlon_dlon;
uniform int   geometry_latlon_Ni;
uniform int   geometry_latlon_Nj;

vec2 getLatLonVertexLonLat (int jglo)
{
  int i = jglo % geometry_latlon_Ni;
  int j = jglo / geometry_latlon_Ni;

  float lat = geometry_latlon_lat0 - geometry_latlon_dlat * float (j);
  float lon = geometry_latlon_lon0 + geometry_latlon_dlon * float (i);

  return vec2 (lon, lat);
}  



// end include geometry/latlon.h


uniform int geometry_type = geometry_none;

vec2 getVertexLonLat (int jglo)
{
  if (geometry_type == geometry_gaussian)
    return getGaussianVertexLonLat (jglo);
  else if (geometry_type == geometry_lambert)
    return getLambertVertexLonLat (jglo);
  else if (geometry_type == geometry_latlon)
    return getLatLonVertexLonLat (jglo);
  else 
    return vertexLonLat;
}
// end include geometry.h

void main ()
{
  vec3 vertexDisp = vec3 (0.0f, 0.0f, 0.0f);

  vec2 vertexLonLat_ = getVertexLonLat (gl_VertexID);

  if (mpiview_scale > 0.0f)
    {
      float lon = vertexMPIView.y, lat = vertexMPIView.z;
      vertexDisp = posFromLonLat (vec2 (lon, lat));
    }

  scalar_vs.fragmentValFlat = vertexVal;
  scalar_vs.fragmentMPIFlat = vertexMPIView.x;

  vec3 vertexPos = posFromLonLat (vertexLonLat_);
  vec3 normedPos = compNormedPos (vertexPos);
  vec3 pos = compProjedPos (vertexPos, normedPos);
  pos = scalePosition (pos, normedPos, scale0);

  if (proj == XYZ)
    {
      pos = pos * (1.0f + height_scale * vertexHeight);
      if (mpiview_scale > 0.0f)
        pos = pos + mpiview_scale * vertexDisp;
    }
    
  gl_Position =  MVP * vec4 (pos, 1.);

  scalar_vs.fragmentVal = vertexVal;
  scalar_vs.fragmentMPI = vertexMPIView.x;
  scalar_vs.fragmentPos = normedPos;
  scalar_vs.missingFlag = vertexVal == 0 ? 1. : 0.;

}
