// start include version.h
#version 440 core
// end include version.h

layout(location = 0) in vec2 vertexLonLat;
layout(location = 1) in float vertexVal_n;
layout(location = 2) in float vertexVal_d;
layout(location = 3) in float vertexHeight;

out 
// start include VECTOR_VS.h
VECTOR_VS
{
  float alpha;
  float fragmentVal;
  vec3 fragmentPos;
} vector_vs;
// end include VECTOR_VS.h

uniform float valmin_n, valmax_n;
uniform float valmin_d, valmax_d;

uniform mat4 MVP;

// start include schmidt.h

uniform mat3 schmidt_rotd;
uniform mat3 schmidt_roti;
uniform float schmidt_opc2 = 1.25;
uniform float schmidt_omc2 = 0.75;
uniform bool schmidt_apply = false;

vec3 applySchmidt (vec3 pos)
{
  if (! schmidt_apply)
    return pos;

  vec3 XYZ = schmidt_roti * pos;
  XYZ = XYZ / length (XYZ);
  
  float lon = atan (XYZ.y, XYZ.x);
  float lat = asin (XYZ.z);
  
  float coordx = lon;
  float coordy = lat;

  float sincoordy = sin (coordy);

  float w = (schmidt_omc2 + sincoordy * schmidt_opc2) 
          / (schmidt_opc2 + sincoordy * schmidt_omc2);
  lat = asin (min (+1.0, max (-1.0, w)));
  lon = coordx;

  float coslat = cos (lat), sinlat = sin (lat);
  float coslon = cos (lon), sinlon = sin (lon);
  
  XYZ = vec3 (coslon * coslat, sinlon * coslat, sinlat);
  XYZ = schmidt_rotd * XYZ;
  XYZ = XYZ / length (XYZ);
  
  return XYZ;
}


// end include schmidt.h
// start include projection.h
// start include projection/kind.h
const int XYZ=0;
const int POLAR_NORTH=1;
const int POLAR_SOUTH=2;
const int MERCATOR=3;
const int LATLON=4;
// end include projection/kind.h

uniform int proj = 0;
uniform bool isflat = true;
uniform float lon0 = 180.0; // Longitude of right handside
const float pi = 3.1415926;

uniform bool apply_coordm = false;
uniform mat3 COORDM;

vec3 posFromLonLat (vec2 vertexLonLat)
{
  float lon = vertexLonLat.x, lat = vertexLonLat.y;
  if (abs (lat) > pi)
    return vec3 (0.0f, 0.0f, 0.0f);
  float coslon = cos (lon), sinlon = sin (lon);
  float coslat = cos (lat), sinlat = sin (lat);
  return vec3 (coslon * coslat, sinlon * coslat, sinlat);
}

vec3 compNormedPos (vec3 vertexPos)
{
  float x = vertexPos.x;
  float y = vertexPos.y;
  float z = vertexPos.z;
  float r = 1. / sqrt (x * x + y * y + z * z);
  return vec3 (x * r, y * r, z * r);
}

vec3 compProjedPos (vec3 vertexPos, vec3 normedPos)
{
  vec3 pos;
  normedPos = applySchmidt (normedPos);
  switch (proj)
    {
      case XYZ:
        if (isflat)
          pos = applySchmidt (normedPos);
        else
          pos = applySchmidt (vertexPos);
        break;
      case POLAR_NORTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (+p.z + 1.0),
                       p.y / (+p.z + 1.0));
        }
        break;
      case POLAR_SOUTH:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          pos =  vec3 (0., p.x / (-p.z + 1.0),
                       p.y / (-p.z + 1.0));
        }
        break;
      case MERCATOR:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = log (tan (pi / 4. + lat / 2.)) / pi;
          pos = vec3 (0., X, Y);
        }
        break;
      case LATLON:
        {
          vec3 p = apply_coordm ? COORDM * normedPos : normedPos;
          float lat = asin (p.z);
          float lon = mod (atan (p.y, p.x), 2 * pi);
          float X = (mod (lon - lon0 * pi / 180.0, 2 * pi) - pi) / pi;
          float Y = lat / pi;
          pos = vec3 (0., X, Y);
        }
        break;
    }

  return pos;
}

// end include projection.h
// start include scale.h
uniform float scale0 = 1.0f;

vec3 scalePosition (vec3 pos, vec3 normedPos, float scale0)
{
  normedPos = applySchmidt (normedPos);
  if (proj == XYZ)
    {
      pos.x = scale0 * normedPos.x;
      pos.y = scale0 * normedPos.y;
      pos.z = scale0 * normedPos.z;
    }
  else if (proj == POLAR_SOUTH)
    {
      pos.x = pos.x - (scale0 - 1.0f);
    }
  else
    {
      pos.x = pos.x + (scale0 - 1.0f);
    }

  return pos;
}

float scalingFactor (vec3 normedPos)
{
  if (proj == POLAR_SOUTH)
    {
      return 1.0 / (1.0 - normedPos.z);
    }
  else if (proj == POLAR_NORTH)
    {
      return 1.0 / (1.0 + normedPos.z);
    }
  else if (proj == MERCATOR)
    {
      return 1.0 / sqrt (1 - normedPos.z * normedPos.z);
    }
  return 1.0;
}

// end include scale.h
// start include unpack.h

uniform int Nmax = 255;

float unpack (float v, float vmin, float vmax)
{
  return vmin + (vmax - vmin) * (Nmax * v - 1) / (Nmax - 1);
}

// end include unpack.h
// start include geometry.h

// start include geometry/buffer_index.h

const int geometry_gaussian_jlat_idx = 1;
const int geometry_gaussian_jglo_idx = 2;
const int geometry_gaussian_glat_idx = 3;


// end include geometry/buffer_index.h
// start include geometry/types.h
const int geometry_none     = 0;
const int geometry_gaussian = 1;
const int geometry_lambert  = 2;
const int geometry_latlon   = 3;

const int dumm_type = geometry_none + geometry_gaussian + geometry_lambert + geometry_latlon;
// end include geometry/types.h
// start include geometry/gaussian.h

// start include geometry/gaussian/kinds.h

const int geometry_gaussian_kind_unk = 0;
const int geometry_gaussian_kind_lin = 1;
const int geometry_gaussian_kind_oct = 2;

const int dumm_kind = geometry_gaussian_kind_unk + geometry_gaussian_kind_lin + geometry_gaussian_kind_oct;

// end include geometry/gaussian/kinds.h

layout (std430, binding=geometry_gaussian_jlat_idx) buffer geometry_gaussian1
{
  int geometry_gaussian_jlat[];
};

layout (std430, binding=geometry_gaussian_jglo_idx) buffer geometry_gaussian2
{
  int geometry_gaussian_jglooff[];
};

layout (std430, binding=geometry_gaussian_glat_idx) buffer geometry_gaussian3
{
  float geometry_gaussian_latgauss[];
};

uniform int   geometry_gaussian_Nj;
uniform mat3  geometry_gaussian_rot;
uniform bool  geometry_gaussian_rotated;
uniform float geometry_gaussian_omc2;
uniform float geometry_gaussian_opc2;
uniform float geometry_gaussian_latfit_coeff[10];
uniform int   geometry_gaussian_latfit_degre;
uniform int   geometry_gaussian_numberOfPoints;
uniform bool  geometry_gaussian_fitlat;
uniform int   geometry_gaussian_kind;

void geometry_gaussian_miss ()
{
}

// start include geometry/gaussian/guess_jlat.h
float geometry_gaussian_guess_lat_lin (int jglo)
{
  return asin (1.0 - 2.0 * float (jglo) / float (geometry_gaussian_numberOfPoints));
}

float geometry_gaussian_guess_lat_oct (int jglo)
{
  return 0.5 * pi * (1.0 - sqrt (2.0 * float (jglo) / float (geometry_gaussian_numberOfPoints)));
}

float geometry_gaussian_guess_lat (int jglo)
{
  float lat = 0.0;
  switch (geometry_gaussian_kind)
    {
      case geometry_gaussian_kind_lin:
        lat = geometry_gaussian_guess_lat_lin (jglo);
        break;
      case geometry_gaussian_kind_oct:
        lat = geometry_gaussian_guess_lat_oct (jglo);
        break;
    }
  return lat;
}


float geometry_gaussian_latfit_eval (int jlat)
{
  float x = float (jlat);
  if (geometry_gaussian_latfit_degre == 4)
    return geometry_gaussian_latfit_coeff[0] + x *
	  (geometry_gaussian_latfit_coeff[1] + x *
	  (geometry_gaussian_latfit_coeff[2] + x *
	  (geometry_gaussian_latfit_coeff[3] + x *
	  (geometry_gaussian_latfit_coeff[4]))));
  else if (geometry_gaussian_latfit_degre == 3)
    return geometry_gaussian_latfit_coeff[0] + x *
	  (geometry_gaussian_latfit_coeff[1] + x *
	  (geometry_gaussian_latfit_coeff[2] + x *
	  (geometry_gaussian_latfit_coeff[3])));
  float y = 0;
  for (int i = geometry_gaussian_latfit_degre; i >= 0; i--)
     y = y * x + geometry_gaussian_latfit_coeff[i];
  return y;
}

int geometry_gaussian_guess_jlat0 (int jglo)
{
  float lat = geometry_gaussian_guess_lat (jglo);
  return int ((0.5 - lat / pi) * (geometry_gaussian_Nj + 1) - 1);
}

int geometry_gaussian_guess_jlat (int jglo)
{
  bool south = jglo > geometry_gaussian_numberOfPoints / 2;
  int iglo = south ? geometry_gaussian_numberOfPoints - 1 - jglo : jglo;

  int ilat0 = geometry_gaussian_guess_jlat0 (iglo);

  // Correction with polynomial
  int ilat1 = int (geometry_gaussian_latfit_eval (ilat0));
  int jlat = ilat0 + ilat1;

  jlat = max (min (jlat, geometry_gaussian_Nj), 0);

  while (iglo < geometry_gaussian_jglooff[jlat])
    {
      jlat = jlat - 1;
      geometry_gaussian_miss ();
    }

  while (iglo >= geometry_gaussian_jglooff[jlat+1])
    {
      jlat = jlat + 1;
      geometry_gaussian_miss ();
    }

  jlat = south ? geometry_gaussian_Nj - 1 - jlat : jlat;

  return jlat;
}


// end include geometry/gaussian/guess_jlat.h

vec2 getGaussianVertexLonLat (int jglo) 
{
  const float twopi = 2.0f * pi;
  int jlat;

  if (geometry_gaussian_fitlat)
    {
      jlat = geometry_gaussian_guess_jlat (jglo);
    }
  else
    {
      jlat = geometry_gaussian_jlat[jglo];
    }

  int jlon = jglo - geometry_gaussian_jglooff[jlat];

  float coordy = geometry_gaussian_latgauss[jlat];
  int pl = geometry_gaussian_jglooff[jlat+1] - geometry_gaussian_jglooff[jlat];
  float coordx = (twopi * float (jlon)) / float (pl);

  float lon, lat;

  if (! geometry_gaussian_rotated)
    {
      lon = coordx;
      lat = coordy;
    }
  else
    {
      float sincoordy = sin (coordy);
      lat = asin ((geometry_gaussian_omc2 + sincoordy * geometry_gaussian_opc2) 
                / (geometry_gaussian_opc2 + sincoordy * geometry_gaussian_omc2));
      lon = coordx;

      float coslat = cos (lat), sinlat = sin (lat);
      float coslon = cos (lon), sinlon = sin (lon);
  
      float X = coslon * coslat;
      float Y = sinlon * coslat;
      float Z =          sinlat;
  
      vec3 XYZ = vec3 (X, Y, Z);
      XYZ = geometry_gaussian_rot * XYZ;
  
      lon = atan (XYZ.y, XYZ.x);
      lat = asin (XYZ.z);
    }


  return vec2 (lon, lat);
}

// end include geometry/gaussian.h
// start include geometry/lambert.h


float sign2 (float a, float b)
{
  return b > 0 ? +abs (a) : -abs (a);
}

struct latlon_t
{
  float lon, lat;
};

struct latlon_j_t
{
  latlon_t V, A, B;
};

struct xy_t
{
  float x, y;
};

struct xy_j_t
{
  xy_t V, A, B;
};

struct rtheta_t
{
  float r, theta;
};

struct rtheta_j_t
{
  rtheta_t V, A, B;
};

float dist_2ref (latlon_t pt_coord, latlon_t ref_coord)
{
  float z = pt_coord.lon - ref_coord.lon;
  z = z - sign2 (pi, z) * (1.0 + sign2 (1.0, abs (z) - pi));
  return -z * sign2 (1.0, z - pi);
}
  
struct proj_t
{
  latlon_t ref_pt;
  float pole;
  float r_equateur;
  float kl;
};

uniform float geometry_lambert_p_pj_ref_pt_lon;
uniform float geometry_lambert_p_pj_ref_pt_lat;
uniform float geometry_lambert_p_pj_pole;
uniform float geometry_lambert_p_pj_r_equateur;
uniform float geometry_lambert_p_pj_kl;
uniform int   geometry_lambert_Nx;
uniform int   geometry_lambert_Ny;
uniform int   geometry_lambert_Nux;
uniform int   geometry_lambert_Nuy;
uniform float geometry_lambert_DxInMetres;
uniform float geometry_lambert_DyInMetres;
uniform float geometry_lambert_center_xy_x;
uniform float geometry_lambert_center_xy_y;

proj_t get_proj_t ()
{
  proj_t proj = proj_t (latlon_t (geometry_lambert_p_pj_ref_pt_lon,
                                  geometry_lambert_p_pj_ref_pt_lat),
                        geometry_lambert_p_pj_pole,
                        geometry_lambert_p_pj_r_equateur,
                        geometry_lambert_p_pj_kl);
  return proj;
}

rtheta_t xy_to_rtheta (proj_t proj, xy_t pt_xy)
{
  float r = sqrt (pt_xy.x * pt_xy.x + pt_xy.y * pt_xy.y);
  float tatng;
  float theta;

  if (pt_xy.y == 0.0) 
    tatng = (pt_xy.x == 0.0) ? pi : sign2 (pi / 2.0, -proj.pole * pt_xy.x);
  else
    tatng = atan (-proj.pole * (pt_xy.x / pt_xy.y));
  theta = pi * sign2 (1.0, pt_xy.x) * (sign2 (0.5, proj.pole * pt_xy.y) + 0.5) + tatng;

  return rtheta_t (r, theta);
}

latlon_t rtheta_to_latlon (proj_t proj, rtheta_t pt_rtheta) 
{
  return latlon_t (proj.ref_pt.lon + pt_rtheta.theta / proj.kl,
                   proj.pole * ((pi / 2.0) - 2.0 * atan (pow (pt_rtheta.r / proj.r_equateur, 1.0 / proj.kl))));
}

latlon_t xy_to_latlon (proj_t proj, xy_t pt_xy) 
{
  return rtheta_to_latlon (proj, xy_to_rtheta (proj, pt_xy));
}

vec2 getLambertVertexLonLat (int jglo)
{
  int i = jglo % geometry_lambert_Nx;
  int j = jglo / geometry_lambert_Nx;

  proj_t proj = get_proj_t ();

  xy_t pt_xy = xy_t ((i - geometry_lambert_Nux / 2) * geometry_lambert_DxInMetres, 
                     (j - geometry_lambert_Nuy / 2) * geometry_lambert_DyInMetres);
  
  pt_xy.x = pt_xy.x + geometry_lambert_center_xy_x;
  pt_xy.y = pt_xy.y + geometry_lambert_center_xy_y;

  latlon_t latlon = xy_to_latlon (proj, pt_xy);

  return vec2 (latlon.lon, latlon.lat);
}  



// end include geometry/lambert.h
// start include geometry/latlon.h
uniform float geometry_latlon_lat0;
uniform float geometry_latlon_lon0;
uniform float geometry_latlon_dlat;
uniform float geometry_latlon_dlon;
uniform int   geometry_latlon_Ni;
uniform int   geometry_latlon_Nj;

vec2 getLatLonVertexLonLat (int jglo)
{
  int i = jglo % geometry_latlon_Ni;
  int j = jglo / geometry_latlon_Ni;

  float lat = geometry_latlon_lat0 - geometry_latlon_dlat * float (j);
  float lon = geometry_latlon_lon0 + geometry_latlon_dlon * float (i);

  return vec2 (lon, lat);
}  



// end include geometry/latlon.h


uniform int geometry_type = geometry_none;

vec2 getVertexLonLat (int jglo)
{
  if (geometry_type == geometry_gaussian)
    return getGaussianVertexLonLat (jglo);
  else if (geometry_type == geometry_lambert)
    return getLambertVertexLonLat (jglo);
  else if (geometry_type == geometry_latlon)
    return getLatLonVertexLonLat (jglo);
  else 
    return vertexLonLat;
}
// end include geometry.h

const float deg2rad = pi / 180.0;

uniform float vscale = 0.01;
uniform bool  arrow_fixed = false;
uniform float head = 0.1;
uniform float posmax = 0.97;

uniform float height_scale = 0.05;
uniform int arrow_kind = 1;
uniform float arrow_min = 0.;

vec3 getPos0 ()
{
  vec3 pos;
       if (gl_VertexID == 0) pos = vec3 (+0.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 1) pos = vec3 (+1.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 2) pos = vec3 (+1.0-head, +head, +0.0);
  else if (gl_VertexID == 3) pos = vec3 (+1.0-head, -head, +0.0);
  else if (gl_VertexID == 4) pos = vec3 (+1.0     ,  +0.0, +0.0);
  return pos;
}

vec3 getPos1 ()
{
  vec3 pos;
       if (gl_VertexID == 0) pos = vec3 (+0.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 1) pos = vec3 (+0.0     , +head, +0.0);
  else if (gl_VertexID == 2) pos = vec3 (+1.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 3) pos = vec3 (+1.0-head, +head, +0.0);
  else if (gl_VertexID == 4) pos = vec3 (+1.0-head, -head, +0.0);
  else if (gl_VertexID == 5) pos = vec3 (+1.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 6) pos = vec3 (+0.0     , -head, +0.0);
  else if (gl_VertexID == 7) pos = vec3 (+0.0     ,  +0.0, +0.0);
  return pos;
}

vec3 getPos2 ()
{
  vec3 pos;
       if (gl_VertexID == 0) pos = vec3 (+0.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 1) pos = vec3 (+0.0     , +head, +0.0);
  else if (gl_VertexID == 2) pos = vec3 (+1.0     ,  +0.0, +0.0);
  else if (gl_VertexID == 3) pos = vec3 (+0.0     , -head, +0.0);
  else if (gl_VertexID == 4) pos = vec3 (+0.0     ,  +0.0, +0.0);
  return pos;
}


uniform float barbxleng;
uniform float barbyleng;
uniform float barbdleng;

uniform float circthres;
uniform int barbthresmax;
uniform float barbthres[10];
uniform int pennthresmax;
uniform float pennthres[10];

vec3 getPos3 (float N, vec2 vertexLonLat)
{
  const int barbvertexmin = 2;
  const int barbvertexmax = barbvertexmin + barbthresmax * 2 - 1;
  const int linevertexmin = barbvertexmax + 1;
  const int linevertexmax = linevertexmin + 1;
  const int pennvertexmin = linevertexmax + 1;
  const int pennvertexmax = pennvertexmin + pennthresmax * 4 - 1;
  const int vertexmax = pennvertexmax;


  vec3 pos = vec3 (0.0f, 0.0f, 0.0f);

//if (gl_InstanceID != 830859)
//  return pos;


  // Circle
  if (N < circthres)
    {
      int k = gl_VertexID;
      k = k + (k % 2);
      float ang = float (k) * pi / (0.5f * (vertexmax - 1));
      pos = barbdleng * vec3 (cos (ang), sin (ang), 0.0f);
      return pos;
    }


  // Number of pennants
  int npenn = 0;

  for (int i = 0; i < pennthresmax + 1; i++)
    if (N >= pennthres[i])
      npenn = i;  
    else
      break;

  N = N - pennthres[npenn];

  // Barb line
  if ((0 <= gl_VertexID) && (gl_VertexID <= 1))
    {
      pos = vec3 (-gl_VertexID, +0.0, +0.0);
    }
  // Barb
  else if ((barbvertexmin <= gl_VertexID) && (gl_VertexID <= barbvertexmax))
    {
      // Barb rank
      int barbind = (gl_VertexID-barbvertexmin) / 2;
      // Point id (=0, or =1)
      int barbext = (gl_VertexID-barbvertexmin) % 2;

      float Nmax = barbthres[barbind+1];
      float Nmin = barbthres[barbind+0];

      // Barb length
      float n = min (max (0, N - Nmin), Nmax - Nmin) / (Nmax - Nmin);

      pos = vec3 (-1.0 + barbind * barbdleng - n * barbxleng * barbext, 
                  n * barbyleng * barbext, +0.0);
    }
  // Extend barb line
  else if ((linevertexmin <= gl_VertexID) && (gl_VertexID <= linevertexmax) && (npenn > 0))
    {
      if (gl_VertexID == 12)
        pos = vec3 (-1.0f, 0.0f, 0.0f);
      else if (gl_VertexID == 13)
        pos = vec3 (-1.0f - barbxleng, 0.0f, 0.0f);
    }
  // Pennants
  else if ((pennvertexmin <= gl_VertexID) && (gl_VertexID <= pennvertexmax))
    {
      // Pennant rank
      int pennind = (gl_VertexID-pennvertexmin) / 4;
      // Point id (0 <= ... <= 3)
      int pennext = (gl_VertexID-pennvertexmin) % 4;

      // If pennant visible
      if (pennind < npenn)
        {
          if (pennext == 0)
            pos = vec3 (-1.0 - (pennind + 0) * barbxleng - barbxleng, 0.0f, 0.0f); 
          else if (pennext == 1)                        
            pos = vec3 (-1.0 - (pennind + 1) * barbxleng - barbxleng, 0.0f, 0.0f); 
          else if (pennext == 2)                        
            pos = vec3 (-1.0 - (pennind + 1) * barbxleng - barbxleng, barbyleng, 0.0f); 
          else if (pennext == 3)                        
            pos = vec3 (-1.0 - (pennind + 1) * barbxleng - barbxleng + barbxleng, 
                        0.0f, 0.0f); 
        }
    }

  pos.y = pos.y * (vertexLonLat.y > 0.0f ? +1 : -1);

  return pos;
}

void main ()
{
  vec2 vertexLonLat_ = getVertexLonLat (gl_InstanceID);
  vec3 vertexPos = posFromLonLat (vertexLonLat_);

  vec3 u = normalize (vec3 (-vertexPos.y, +vertexPos.x, 0.));
  vec3 v = cross (vertexPos, u);

  vector_vs.fragmentVal = vertexVal_n;
  vector_vs.fragmentPos = vertexPos;

  float N = unpack (vertexVal_n, valmin_n, valmax_n);
  float D = unpack (vertexVal_d, valmin_d, valmax_d);

  bool defined = (vertexVal_d != 0) && (N > arrow_min);
//defined = gl_InstanceID == 830859;
  vec3 pos;

  if (! defined)
    pos = vec3 (+0.0, +0.0, +0.0);
  else if (arrow_kind == 0)
    pos = getPos0 ();
  else if (arrow_kind == 1)
    pos = getPos1 ();
  else if (arrow_kind == 2)
    pos = getPos2 ();
  else if (arrow_kind == 3)
    pos = getPos3 (N, vertexLonLat_);
  else 
    pos = getPos0 ();
    

  vector_vs.alpha = 1.;
  if (defined)
    {
      D = D * deg2rad;

      float X, Y;

      if (arrow_fixed)
        {
          X = vscale * cos (D);
          Y = vscale * sin (D);
        }
      else
        {
          X = vscale * N * cos (D) / valmax_n;
          Y = vscale * N * sin (D) / valmax_n;
        }
     
      float s = scalingFactor (compNormedPos (vertexPos));

      pos = vertexPos + ((pos.x * X - pos.y * Y) * u + (pos.x * Y + pos.y * X) * v) / s;
     
      vec3 normedPos = compNormedPos (pos);
      vec3 projedPos = compProjedPos (pos, normedPos);
      pos = scalePosition (projedPos, normedPos, scale0);

      if (proj == XYZ)
        pos = pos * (1.0f + height_scale * vertexHeight);

      if ((proj == LATLON) || (proj == MERCATOR))
      if ((pos.y < -posmax) || (+posmax < pos.y))
        {
          pos.x = -0.1;
          vector_vs.alpha = 0.0;
	}
    }

  gl_Position =  MVP * vec4 (pos, 1.);

}
